/*----------------------------------------------------------------------------
    TAPESTREA: Techniques And Paradigms for Expressive Synthesis, 
               Transformation, and Rendering of Environmental Audio
      Engine and User Interface

    Copyright (c) 2006 Ananya Misra, Perry R. Cook, and Ge Wang.
      http://taps.cs.princeton.edu/
      http://soundlab.cs.princeton.edu/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: ui_audio.cpp
// desc: birdbrain ui audio
//
// authors: Ananya Misra (amisra@cs.princeton.edu)
//          Ge Wang (gewang@cs.princeton.edu)
//          Perry R. Cook (prc@cs.princeton.edu)
// date: Winter 2004
//-----------------------------------------------------------------------------
#ifndef __UI_AUDIO_H__
#define __UI_AUDIO_H__

#include "taps_birdbrain.h"
#include "audicle_def.h"
#include "rtaudio.h"
#include "util_thread.h"
#include "ui_audiofx.h"

// libsndfile
#ifndef __USE_SNDFILE_PRECONF__
#include <sndfile.h>
#else
#include "util_sndfile.h"
#endif


// forward reference
class TreesynthIO;


//-----------------------------------------------------------------------------
// name: AudioSrc
// desc: ...
//-----------------------------------------------------------------------------
struct AudioSrc
{
    AudioSrc() { m_done = TRUE; m_stop_asap = FALSE; on = NULL; m_delete = TRUE; 
                 m_gain = 1.0f; m_pan = .5f; }
    virtual ~AudioSrc() { m_done = TRUE; if( on ) *on = FALSE; }
    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames ) = 0;
    virtual void reset() { m_done = FALSE; m_stop_asap = FALSE; }
    virtual void rewind() { }
    virtual t_TAPBOOL done() { return m_done; }
    virtual t_TAPBOOL playing() { return !m_done; }
    virtual const std::string & name() { return m_name; }
    std::string m_name;

    virtual void set_gain( t_TAPSINGLE gain )
    { if( gain<0 ) gain = 0; if( gain>128 ) gain = 128; m_gain = gain; }
    virtual t_TAPSINGLE get_gain() const { return m_gain; }
    virtual void set_pan( t_TAPSINGLE pan )
    { if( pan<0 ) pan = 0; if( pan>1 ) pan = 1; m_pan = pan; }
    virtual t_TAPSINGLE get_pan() const { return m_pan; }
    
    t_TAPBOOL * on; 
    t_TAPBOOL m_done; // whether it's done (safe to delete) (not in use) (...)
    t_TAPBOOL m_stop_asap; // (ignore: whether it's not done) (the real comment: whether to stop asap ) 
    t_TAPBOOL m_delete;
    SAMPLE m_gain;
    SAMPLE m_pan;

    static void mono2stereo( SAMPLE * self, t_TAPUINT frames );
    static void stereo2mono( SAMPLE * self, t_TAPUINT frames );
    static void mono2stereo( const SAMPLE * mono, t_TAPUINT frames, SAMPLE * stereo );
    static void stereo2mono( const SAMPLE * stereo, t_TAPUINT frames, SAMPLE * mono );
    static void pan_buf( SAMPLE * stereo, t_TAPUINT frames, t_TAPSINGLE where );
    static void gain_buf( SAMPLE * buffer, t_TAPUINT num_samples, SAMPLE gain );
};




//-----------------------------------------------------------------------------
// name: AudioSrcTest
// desc: ...
//-----------------------------------------------------------------------------
struct AudioSrcTest : public AudioSrc
{
    AudioSrcTest( t_TAPFLOAT freq ) : AudioSrc() { m_freq = freq; m_t = 0.0; }
    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );

protected:
    t_TAPFLOAT m_freq;
    t_TAPFLOAT m_t;
};




//-----------------------------------------------------------------------------
// name: AudioSrcFile
// desc: ...
//-----------------------------------------------------------------------------
struct AudioSrcFile : public AudioSrc
{
    AudioSrcFile( );
    virtual ~AudioSrcFile();

    virtual t_TAPBOOL mtick( SAMPLE * buffer, t_TAPUINT num_samples );
    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );
    virtual void rewind() { AudioSrc::rewind(); sf_seek( sf, 0, SEEK_SET ); }
    t_TAPBOOL open( const char * filename, t_TAPUINT start = 0, t_TAPUINT len = 0, t_TAPBOOL scale = FALSE );
    t_TAPBOOL close();
    t_TAPBOOL seek( t_TAPUINT where, t_TAPUINT how )
    { return sf_seek( sf, where, how ) < sf_info.frames; }
    const std::string & filename() { return m_filename; }
    const std::string & last_error() { return m_lasterror; }
    const SF_INFO & info() { return sf_info; }
    t_TAPUINT start() { return m_start; }
    t_TAPUINT end() { return m_end; }
    t_TAPUINT now() { return sf_seek( sf, 0, SEEK_CUR ); }

    t_TAPSINGLE * slide;
    bool * slide_locally;

protected:
    SAMPLE * m_buffer;
    t_TAPUINT m_buffer_size;
    SNDFILE * sf;
    SF_INFO sf_info;
    t_TAPUINT m_start;
    t_TAPUINT m_end;
    std::string m_filename;
    t_TAPSINGLE signal_scale;
    std::string m_lasterror;
};




//-----------------------------------------------------------------------------
// name: AudioSrcFrame
// desc: ...
//-----------------------------------------------------------------------------
struct AudioSrcFrame : public AudioSrc
{
    AudioSrcFrame( Frame * f ) : AudioSrc() { frame = f; pos = 0; slide = NULL; slide_locally = NULL; }
    virtual t_TAPBOOL mtick( SAMPLE * buffer, t_TAPUINT num_samples ); 
    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );
    virtual void rewind() { AudioSrc::rewind(); pos = 0; }
    t_TAPBOOL seek( t_TAPUINT where, t_TAPUINT how );
    int num_samples();

    t_TAPSINGLE * slide;
    bool * slide_locally;

protected:
    Frame * frame;
    t_TAPUINT pos;
};




//-----------------------------------------------------------------------------
// name: AudioSrcEliot
// desc: ...
//-----------------------------------------------------------------------------
struct AudioSrcEliot : public AudioSrc
{
    AudioSrcEliot( TreesynthIO * io ) : AudioSrc() { ts_io = io; }
    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );
    void stop() { m_stop_asap = TRUE; }

protected:
    TreesynthIO * ts_io;
};




//-----------------------------------------------------------------------------
// name: AudioFxReverb
// desc: ...
//-----------------------------------------------------------------------------
struct AudioFxReverb : public AudioSrc
{
public:
    AudioFxReverb() : AudioSrc()
    { m_mix = 0; m_state = NONE; m_left = m_right = NULL; }
    virtual ~AudioFxReverb()
    { SAFE_DELETE( m_left ); SAFE_DELETE( m_right ); m_state = NONE; }

    enum { NONE, PRC, JC, N };
    virtual void set( t_TAPUINT which, t_TAPFLOAT T60 );
    virtual void mix( t_TAPFLOAT value );
    t_TAPUINT getset() const { return m_state; }
    t_TAPFLOAT getmix() const { return m_mix; }

    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );

protected:
    t_TAPFLOAT m_mix;
    t_TAPUINT m_state;
    FxReverb * m_left;
    FxReverb * m_right;
};




//-----------------------------------------------------------------------------
// name: AudioBus
// desc: ...
//-----------------------------------------------------------------------------
struct AudioBus : public AudioSrc
{
    AudioBus( t_TAPUINT num_frames ) : AudioSrc() { this->init( num_frames ); }
    virtual ~AudioBus();

    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );

    virtual void play( AudioSrc * src, t_TAPBOOL solo = FALSE, t_TAPBOOL rew = TRUE );
    virtual void remove( AudioSrc * src );
    virtual void stop();

    virtual t_TAPUINT num_src() const { return m_src.size(); }

    AudioFxReverb * reverb() { return &m_reverb; }
    AudioFxReverb m_reverb;

    t_TAPBOOL m_ready;
    std::vector<AudioSrc *> m_src;
    std::vector<AudioSrc *> m_src2;
    SAMPLE * m_buffer;
    SAMPLE * m_buffer2;
    t_TAPUINT m_num_frames;
    XMutex m_mutex;

private:
    t_TAPBOOL init( t_TAPUINT buffer_size );
};




//-----------------------------------------------------------------------------
// name: AudioBusParallel
// desc: ...
//-----------------------------------------------------------------------------
struct AudioBusParallel : public AudioBus
{
    AudioBusParallel( t_TAPUINT num_frames, t_TAPUINT sig_frames );
    virtual ~AudioBusParallel() { close(); }
    virtual t_TAPBOOL stick( SAMPLE * buffer, t_TAPUINT num_frames );
    void accumulate();

    bool open( const std::string & filename );
    void close();
    SNDFILE * sf;
    SF_INFO sf_info;
    t_TAPUINT m_sig_frames;

protected:
    XThread m_thread;
    XMutex m_sf_mutex;
};




//-----------------------------------------------------------------------------
// name: AudioCentral
// desc: ...
//-----------------------------------------------------------------------------
struct AudioCentral
{
    virtual ~AudioCentral() { if( m_audio ) this->free(); }
    static AudioCentral * instance();

    t_TAPBOOL init( t_TAPUINT srate, t_TAPUINT num_frames,
                   t_TAPUINT sig_frames, t_TAPUINT bus = 8 );
    t_TAPBOOL free();

    virtual t_TAPBOOL stop_all();
    AudioBus * bus( t_TAPUINT index );
    virtual t_TAPUINT num_bus() const { return m_bus->m_src.size(); }

    virtual void set_gain( t_TAPSINGLE gain ) { m_bus->set_gain( gain ); }
    virtual t_TAPSINGLE get_gain() const { return m_bus->get_gain(); }

    t_TAPBOOL record_start( const std::string & filename )
    { return m_bus->open( filename ); }
    t_TAPBOOL record_stop()
    { m_bus->close(); return TRUE; }

public:
    SAMPLE * m_buffer;
    SAMPLE * m_buffer2;
    t_TAPUINT m_num_frames;
    t_TAPUINT m_sig_frames;
    t_TAPUINT m_channels;
    t_TAPUINT m_srate;
    RtAudio * m_audio;
    XMutex m_mutex;
    AudioBusParallel * m_bus;

protected:
    static AudioCentral * our_instance;
    AudioCentral() { m_buffer = NULL; m_buffer2 = NULL; m_num_frames = 0;
                     m_srate = 0; m_audio = NULL; m_bus = NULL; m_channels = 2; }
};




#endif
